#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include "constants.h"
#include "atree.h"

static const char *rcsid = "$Id: atree.c,v 1.7 1994/05/13 14:05:53 queinnec Exp $";

#define ECONOMICALLOC 1

const char *TreeCompl_SuccStr[] = {
    "Not done", "Fail", "Exact", "Exact but longer", "Partial"
};

typedef struct Node Node;

struct Node {
    char   thechar;   /* thechar == 0 means unused */
    char   endofword; /* 1 if a word ends here */
#if ECONOMICALLOC
    char   first;     /* Is it the first Node of a alloc block? */
#endif
    Node  *next;
    Node  *sons;
    Node  *father;
};

struct Tree {
    Node   root; /* A fake node, which is a father for every node. */
    short  caseinsensitive;
};

#if !ECONOMICALLOC

#  define mallocnode() malloc(sizeof(Node))
#  define freenode(n)  free((n))

#else /* ECONOMICALLOC */

#define UNITOFALLOC 20

static Node *mallocnode (void)
{
    static int left = 0;
    static Node *next = NULL;
    Node *r;
    if (left == 0) {
        Node *n;
        int i;
        next = (Node *) malloc (sizeof (Node) * UNITOFALLOC);
        if (next == NULL) {
            fprintf (stderr, "Tree: malloc node: malloc failed\n");
            return (NULL);
        }
        left = UNITOFALLOC;
        for (n = next, i = UNITOFALLOC; i != 0; i--, n++) {
            n->thechar = 0;
            n->first = 0;
        }
        next->first = 1;
    }
    left--;
    r = next;
    next = &(next[1]);
    return (r);
}

static void freenode (Node *n)
{
    int i;
    Node *nn;
    int unused;
    n->thechar = 0;
    while (n->first != 1)
      n--;
    unused = 1;
    for (nn = n, i = UNITOFALLOC; i != 0; i--, nn++) {
        if (nn->thechar != 0) {
            unused = 0;
            break;
        }
    }
    if (unused)
      free (n);
}

#endif /* ECONOMICALLOC */

static Node *createnode (Tree *tree, Node *father, char c)
{
    Node *n;
    assert (tree != NULL);
    n = mallocnode ();
    if (n == NULL)
      return (NULL);
    n->endofword = FALSE;
    if (tree->caseinsensitive)
      c = (isupper (c) ? tolower(c) : c);
    n->thechar = c;
    n->father = father;
    if (father != NULL) {
        n->next = father->sons;
        father->sons = n;
    }
    n->sons = NULL;
    return (n);
}

static Node *getnode (Tree *tree, Node *node, char c, int create)
{
    Node *n;
    assert (tree != NULL);
    assert (node != NULL);
    if (tree->caseinsensitive)
      c = (isupper (c) ? tolower(c) : c);
    n = node->sons;
    while ((n != NULL) && (n->thechar != c)) {
        n = n->next;
    }
    if ((n == NULL) && create) {
        n = createnode (tree, node, c);
    }
    return (n);
}



Tree *tree_create (int caseinsensitive)
{
    Tree *r;
    r = (Tree *) malloc (sizeof (Tree));
    if (r == NULL) {
        fprintf (stderr, "Tree: create tree: malloc failed\n");
        return (NULL);
    }
    r->caseinsensitive = caseinsensitive;
    r->root.thechar = 0;
    r->root.endofword = FALSE;
    r->root.next = NULL;
    r->root.sons = NULL;
    r->root.father = NULL;
    return (r);
}

static void destroynode (Node *n)
{
    Node *son;
    assert (n != NULL);
    son = n->sons;
    while (son != NULL) {
        Node *next = son->next;
        destroynode (son);
        son = next;
    }
    freenode (n);
}

int   tree_destroy (Tree *tree)
{
    Node *n;
    n = tree->root.sons;
    while (n != NULL) {
        Node *new = n->next;
        destroynode (n);
        n = new;
    }
    free (tree);
    return (OK);
}

int   tree_insert (Tree *tree, const char *word)
{
    const char *p = word;
    Node *n;
    assert (tree != NULL);
    if (*p == 0)
      return (ERROR);
    n = &(tree->root);
    for (; *p != 0; p++) {
        n = getnode (tree, n, *p, TRUE);
        if (n == NULL)
          return (ERROR);
    }
    n->endofword = TRUE;
    return (OK);
}

int   tree_delete (Tree *tree, const char *word)
{
    Node *n;
    if (*word == 0)
      return (ERROR);
    n = &(tree->root);
    for (; *word != 0; word++) {
        n = getnode (tree, n, *word, FALSE);
        if (n == NULL)
          return (ERROR);
    }
    if (! n->endofword)
      return (ERROR);
    n->endofword = FALSE;
    do {
        Node *f;
        Node *nn;
        if (n->sons != NULL)
          return (OK);
        if (n->endofword)
          return (OK);
        /* delete the node */
        f = n->father;
        if (f->sons == n) {
            f->sons = n->next;
        } else {
            for (nn = f->sons; nn->next != n; nn = nn->next)
              /* loop */;
            nn->next = n->next;
        }
        freenode (n);
        n = f;
    } while (n != &(tree->root));
    return (OK);
}

int   tree_check (Tree *tree, const char *word)
{
    const char *p = word;
    Node *n;
    if (*p == 0)
      return (FALSE);
    n = &(tree->root);
    for (; *p != 0; p++) {
        n = getnode (tree, n, *p, FALSE);
        if (n == NULL)
          return (FALSE);
    }
    if (n->endofword)
      return (TRUE);
    else
      return (FALSE);
}




/********************* COMPLETION ********************/

struct TreeCompletion {
    Tree             *tree;
    char             *w;
    Node             *node;    /* The node which holds the last char of w */ 
    int               number;  /* -1 means not yet computed */
    int               len;     /* len of w */
    Node             *last;    /* last node done for getnextcompletion */
    TreeCompl_Success success;
};

TreeCompletion *tree_createcompletion (Tree *tree)
{
    TreeCompletion *r;
    r = (TreeCompletion *) malloc (sizeof (TreeCompletion));
    if (r == NULL) {
        fprintf (stderr, "Tree: create completion: malloc failed\n");
        return (NULL);
    }
    r->tree = tree;
    r->w = NULL;
    r->number = -1;
    r->node = NULL;
    r->len = 0;
    r->success = TREECOMPL_NOTDONE;
    r->last = NULL;
    return (r);
}

int tree_destroycompletion (TreeCompletion *tc)
{
    if (tc == NULL)
      return (ERROR);
    if (tc->w != NULL)
      free (tc->w);
    free (tc);
    return (OK);
}

TreeCompl_Success tree_trycompletion (TreeCompletion *tc, const char *word)
{
    const char *p = word;
    Node *n;
    Tree *tree;
    assert (tc != NULL);
    if (tc->w != NULL)
      free (tc->w);
    tc->last = NULL;
    if (*p == 0) {
        tc->success = TREECOMPL_FAIL;
        return (tc->success);
    }
    tree = tc->tree;
    tc->w = (char *) strdup (word);
    tc->len = strlen (word);
    tc->number = -1;
    n = &(tree->root);
    for (; *p != 0; p++) {
        n = getnode (tree, n, *p, FALSE);
        if (n == NULL) {
            tc->success = TREECOMPL_FAIL;
            break;
        }
    }
    if (n != NULL) {
        if (n->endofword)
          tc->success = TREECOMPL_EXACT;
        else
          tc->success = TREECOMPL_PARTIAL;
    }
    tc->node = n;
    if (tc->success == TREECOMPL_EXACT) {
        if (n->sons != NULL) {
            /* si sons est non NULL, c'est qu'un mot plus long existe */
            tc->success = TREECOMPL_EXACTBUTLONGER;
        }
    }
    return (tc->success);
}

static int nodecount (Node *n)
{
    Node *son;
    int r = 0;
    assert (n != NULL);
    if (n->endofword)
      r = 1;
    for (son = n->sons; son != NULL; son = son->next) {
        r += nodecount (son);
    }
    return (r);
}

int tree_nbofcompletions (TreeCompletion *tc)
{
    assert (tc != NULL);
    if (tc->success == TREECOMPL_NOTDONE)
      return (-1);
    if (tc->number != -1)
      return (tc->number);
    if (tc->success == TREECOMPL_FAIL) {
        tc->number = 0;
        return (tc->number);
    }
    tc->number = nodecount (tc->node);
    return (tc->number);
}

int partialcount (Node *n, int *currentcount, int max)
{
    Node *son;
    if (n == NULL)
      return (0);
    if (n->endofword)
      ++*currentcount;
    if (*currentcount >= max)
      return (TRUE);
    for (son = n->sons; son != NULL; son = son->next) {
        if (partialcount (son, currentcount, max))
          return (TRUE);
    }
    return (FALSE);
}

int tree_nbofcompletionshigherthan (TreeCompletion *tc, int max)
{
    int num = 0;
    assert (tc != NULL);
    if (tc->success == TREECOMPL_NOTDONE)
      return (FALSE);
    if (tc->number != -1)
      return (tc->number >= max);
    if (tc->success == TREECOMPL_FAIL) {
        tc->number = 0;
        return (FALSE);
    }
    return (partialcount (tc->node, &num, max));
 }

static char *buildstringwithlen (Node *n, int len)
{
    char *s = (char *) malloc (len + 1);
    char *ss;
    if (s == NULL) {
        fprintf (stderr, "Tree: build completion: malloc failed\n");
        return (NULL);
    }
    ss = &s[len - 1];
    s[len] = 0;
    while (n->thechar != 0) {
        *ss = n->thechar;
        n = n->father;
        ss--;
    }
    return (s);
}

static char *buildstring (Node *n)
{
    int len = 0;
    Node *nn = n;
    while (nn->thechar != 0) {
        len++;
        nn = nn->father;
    }
    return (buildstringwithlen (n, len));
}

static int nodeaction (Node *n, int len, forallcompletion forall, void *info)
{
    Node *son;
    if (n == NULL)
      return (FALSE);
    if (n->endofword) {
        char *s = buildstringwithlen (n , len);
        if (s == NULL)
          return (TRUE);
        if (forall (s, info))
          return (TRUE);
    }
    for (son = n->sons; son != NULL; son = son->next) {
        if (nodeaction (son, len + 1, forall, info))
          return (TRUE);
    }
        
    return (FALSE);
}

int tree_forallcompletion (TreeCompletion *tc, forallcompletion forall,
                           void *info)
{
    assert (tc != NULL);
    if (tc->success == TREECOMPL_NOTDONE)
      return (ERROR);
    if (tc->success == TREECOMPL_FAIL)
      return (ERROR);
    nodeaction (tc->node, tc->len, forall, info);
    return (OK);
}

static int findfirst (char *word, void *info)
{
    char **infos = (char **)info;
    *infos = word;
    return (TRUE);
}

char *tree_getcompletion (TreeCompletion *tc)
{
    char *s;
    assert (tc != NULL);
    if (tc->success == TREECOMPL_NOTDONE)
      return (NULL);
    if (tc->success == TREECOMPL_FAIL)
      return (NULL);
    tree_forallcompletion (tc, findfirst, &s);
    return (s);
}

struct holder {
    int match_list_size;
    int matches;
    char **match_list;
};

static int addcompletion (char *w, void *info)
{
    struct holder *m = (struct holder *)info;
    if (m->matches + 1 == m->match_list_size) {
        m->match_list_size += 10;
        m->match_list = (char **) realloc (m->match_list,
                                 (m->match_list_size + 1) * sizeof (char *));
        if (m->match_list == NULL) {
            fprintf (stderr, "Tree: build all completions: realloc failed\n");
            return (TRUE);
        }
    }
    m->match_list[m->matches++] = w;
    m->match_list[m->matches] = (char *)NULL;
    return (FALSE);
}

char **tree_getallcompletions (TreeCompletion *tc, int *howmany)
{
    struct holder holder;
    if ((tc->success == TREECOMPL_NOTDONE) || (tc->success == TREECOMPL_FAIL))
      return (NULL);
    holder.matches = 0;
    holder.match_list_size = 10;
    holder.match_list = (char **) malloc ((holder.match_list_size + 1) * sizeof (char *));
    if (holder.match_list == NULL) {
        fprintf (stderr, "Tree: build all completions: realloc failed\n");
        return (NULL);
    }
    holder.match_list[0] = NULL;
    tree_forallcompletion (tc, addcompletion, &holder);
    *howmany = holder.matches;
    return (holder.match_list);
}

char *tree_getnextcompletion (TreeCompletion *tc)
{
    Node *current;
    Node *father;
    if ((tc->success == TREECOMPL_NOTDONE) || (tc->success == TREECOMPL_FAIL))
      return (NULL);
    if (tc->last == NULL) {
        tc->last = tc->node;
        if (tc->node->endofword) {
            return (buildstring (tc->node));
        }
    }
    current = tc->last->sons;
    father = tc->last;
    while (1) {
        while (current != NULL) {
            if (current->endofword) {
                /* We have found one word! Stop here,
                 * and remember where we get */
                tc->last = current;
                return (buildstring (current));
            }
            /* No word, go deeper in the tree. */
            current = current->sons;
            father = current;
        }
        /* current == NULL, so we are at a leaf.
         * We just have to climp up, and go to the next subtree.
         * However, we must not go higher than the initial node. */
        if (father == tc->node)
          /* When we are back at the starting node, it means we have
           * visited all the subtree. */
          return (NULL);
        current = father->next;
        father = father->father;
    } /* while */
}



/********************* DEBUG ********************/

static void dumpnode (Node *n, int indent)
{
    Node *son;
    if (n == NULL)
      return;
    printf ("%*cchar: %c, eow: %d\n", indent, ' ',
            n->thechar, n->endofword);
    for (son = n->sons; son != NULL; son = son->next) {
        dumpnode (son, indent + 1);
    }
}

void tree_dump (Tree *tree)
{
    Node *son;
    printf ("Dumping...\n");
    for (son = tree->root.sons; son != NULL; son = son->next)
      dumpnode (son, 1);
    printf ("dump end\n");
}


typedef struct {
    int nbofnodes;
    int nbofwords;
    int nbofuseless;
    int maxnboflinked;
} Stat;

void statnode (Node *n, Stat *treestat)
{
    Node *son;
    int nb = 0;
    assert (n != NULL);
    treestat->nbofnodes++;
    if (n->endofword)
      treestat->nbofwords++;
    for (son = n->sons; son != NULL; son = son->next) {
        nb++;
        statnode (son, treestat);
    }
    if (nb > treestat->maxnboflinked)
      treestat->maxnboflinked = nb;
    if ((!n->endofword) && (n->sons == NULL))
      treestat->nbofuseless++;
}

void tree_stat (Tree *tree)
{
    Stat treestat;
    Node *son;
    int nb;
    treestat.nbofnodes = 0;
    treestat.nbofwords = 0;
    treestat.nbofuseless = 0;
    treestat.maxnboflinked = 0;
    printf ("Statistics:\n");
    nb = 0;
    for (son = tree->root.sons; son != NULL; son = son->next) {
        nb++;
        statnode (son, &treestat);
    }
    if (nb > treestat.maxnboflinked)
      treestat.maxnboflinked = nb;
    printf (" Nb of nodes: %d\n", treestat.nbofnodes);
    printf (" Nb of words: %d\n", treestat.nbofwords);
    printf (" Nb of useless nodes: %d\n", treestat.nbofuseless);
    if (treestat.nbofwords != 0)
      printf (" Average word length: %.1f\n",
              ((double) treestat.nbofnodes) / treestat.nbofwords);
    printf (" Max nb of linked nodes: %d\n", treestat.maxnboflinked);
}
