/* $Id: atree.h,v 1.2 1994/03/17 13:14:57 queinnec Exp $ */

/*
 * Package de manipulation d'arbres pour rechercher des mots.
 *
 * Il est possible d'inserer un mot, de verifier qu'il existe, et
 * d'obtenir la liste des mots commencant par un prefixe (completion)
 *
 * On ne peut pas detruire des mots alors qu'on a des completions en court.
 * On peut en rajouter, mais je ne sais pas trop ce qu'il se passe.
 */

#ifndef _ATREE_H
#define _ATREE_H

typedef struct Tree Tree;

/*
 * Les caracteres standards (actuellement a-z).
 */
extern const char *normalchars;

/*
 * Creation d'un arbre.
 */
Tree *tree_create (int caseinsensitive);

/*
 * Destruction de l'arbre et des structures associees.
 */
int   tree_destroy (Tree *);

/*
 * Insertion d'un mot.
 */
int   tree_insert (Tree *, const char *word);

/*
 * Destruction d'un mot.
 * NOT YET IMPLEMENTED
 */
int   tree_delete (Tree *, const char *word);

/*
 * Verification d'un mot.
 */
int   tree_check (Tree *, const char *word);


/* Completion */

/*
 * Le premier  principe d'utilisation est:
    tc = tree_createcompletion (tree);
    success = tree_trycompletion (tc, word);
    if (success == TREECOMPL_FAIL) {
       echec
    } else if (success == TREECOMPL_EXACT) {
       exact
    } else if (tree_nbofcompletionshigherthan (tc, maxallowed)) {
       too many completion
    } else {
       num = tree_nbofcompletions (tc);
       if (num == 1) {
          one completion
          wordcomplete = tree_getcompletion (tc);
       } else {
          tree_forallcompletion (tc, dosomething, info);
       }
    }
 *
 * Le second est tout betement:
    tc = tree_createcompletion (tree);
    success = tree_trycompletion (tc, word);
    while ((s = tree_getnextcompletion (tc)) != NULL) {
       do anything with it.
       free (s);
    }
 *
 * Il y a aussi tree_getallcompletions qui renvoie un char**, mais c'est
 * historique plus qu'autre chose.
 */

typedef struct TreeCompletion TreeCompletion;

typedef enum {
    TREECOMPL_NOTDONE,
    TREECOMPL_FAIL,           /* No completion */
    TREECOMPL_EXACT,          /* The word is the only valid word */
    TREECOMPL_EXACTBUTLONGER, /* The word is valid, but other longer words exist */
    TREECOMPL_PARTIAL         /* Only longer words exist. */
} TreeCompl_Success;

extern const char *TreeCompl_SuccStr[]; /* For debug */

TreeCompletion *tree_createcompletion (Tree *);
int             tree_destroycompletion (TreeCompletion *);

TreeCompl_Success tree_trycompletion (TreeCompletion *, const char *word);

int tree_nbofcompletions (TreeCompletion *);

/*
 * Returns TRUE if there is max or more completions.
 */
int tree_nbofcompletionshigherthan (TreeCompletion *, int max);

/*
 * Returns the first completion of the word.
 * To be used (preferably) if there is only one completion.
 * (I can't understand its use otherwise).
 */
char *tree_getcompletion (TreeCompletion *);

/*
 * For all completions, call the function with a valid completion.
 * The word is malloced and IS NOT FREED.
 * tree_forallcompletion stops as soon as forallcompletion returns TRUE.
 */
typedef int (*forallcompletion) (char *word, void *info);

int tree_forallcompletion (TreeCompletion *, forallcompletion, void *);

/*
 * get all the completions, and put them in a array terminated with NULL.
 * howmany also holds the number of found completions.
 * Everything is malloc'd. The user must free the data itself.
 */
char **tree_getallcompletions (TreeCompletion *, int *howmany);

/*
 * return all the completion, one after the other, and return NULL when
 * there is no more string.
 * Every string is malloc'd. The user must free the data itself.
 */
char *tree_getnextcompletion (TreeCompletion *);

/* Debug */

void tree_stat (Tree *tree);

void tree_dump (Tree *tree);

#endif /* _ATREE_H */
