#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <malloc.h>
#include <string.h>
#if HAVE_VFORK && defined(sparc)
#  include <vfork.h>
#endif
#include "moviedb.h" /* only for dbutils.h */
#include "dbutils.h" /* only for duplicateString */
#include "constants.h"
#include "exec.h"

static const char *rcsid = "$Id";

#if HAVE_VFORK
#  define fork() (vfork())
#endif

/*
 * The default output parameters, set by set_default_ouput() and used by
 * restore_default_output().
 */
static int defaultoutput = REDIR_CMDE;
static char *defaultoutputname = NULL;

/* This is the current output as set by set_command_output. */
static int   output = REDIR_CMDE;
static char *outputname = NULL;

/*
 * Set the default output. Should be called only once.
 * (well, you can call it more than once, but without effect).
 *
 * If name is NULL, set the default values of the default output (ouch!).
 * set_default_output (anything, NULL) can so be used to initialize the output
 *  variables with their default value.
 */
void set_default_output (Redirection redir, const char *name)
{
    if (defaultoutputname != NULL)
      return;
    if (name == NULL) {
        defaultoutput = REDIR_CMDE;
        name = (char *) getenv ("PAGER");
        if (name == NULL)
          name = MORE;
    } else {
        defaultoutput = redir;
    }
    defaultoutputname = duplicateString (name);
    /* and then set the current output. */
    set_command_output (defaultoutput, defaultoutputname);
}

void restore_default_output (void)
{
    if (defaultoutputname == NULL) {
        /* PANIC!! nothing has been initialized! Do it *now*! */
        set_default_output (0, NULL);
        return;
    }
    set_command_output (defaultoutput, defaultoutputname);
}

void set_command_output (Redirection redir, const char *name)
{
    set_default_output (0, NULL); /* ensure we have been initialized */
    if (name == NULL) {
        /* aaarrrggghhh, what does this means? */
        return;
    }
    output = redir;
    if (outputname != NULL)
      free (outputname);
    outputname = duplicateString (name);
}

/*
 * Execute: cmde opt w "-full"
 * The output either goes into a file or is piped into a command, based on
 *  outputtofile.
 * If outputtofile is FALSE:
 *    if outputname contains at least a space, sh -c outputname is used,
 *    else exec outputname is directly used.
 * If outputtofile is TRUE: output goes into a file.
 * This means that setting PAGER to "cat >/tmp/toto" or "tee /tmp/toto |more"
 *  works!
 */
void docommand (const char *w, const char *cmde, const char *opt)
{
    int pid;
    set_default_output (FALSE, NULL); /* ensure we have been initialized */
    pid = fork();
    if (pid == -1) {
        perror ("Cannot fork");
        return;
    } else if (pid == 0) {
        /* Child side. */
        if (output != REDIR_CMDE) {
            /* Output goes into a file. */
            int fd;
            if (output == REDIR_FILE)
              fd = open (outputname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
            else                /* REDIR_APPEND */
              fd = open (outputname, O_WRONLY|O_CREAT|O_APPEND, 0666);
            if (fd == -1) {
                perror (outputname);
                _exit (1);
            }
            close (1);
            dup (fd);
            if (execlp (cmde, cmde, opt, w, "-full", NULL) == -1) {
                perror (cmde);
                exit (1);
            }
        } else {
            /* Output goes into a command. */
            int fd[2];
            if (pipe (fd) == -1) {
                perror ("Cannot create pipe");
                _exit (1);
            }
            pid = fork ();
            if (pid == -1) {
                perror ("Cannot fork");
                close (fd[0]);
                close (fd[1]);
                _exit (1);
            } else if (pid == 0) {
                /* Child side. Execute the command. */
                close(1);
                dup (fd[1]);
                close (fd[0]);
                if (execlp (cmde, cmde, opt, w, "-full", NULL) == -1) {
                    perror (cmde);
                    exit (1);
                }
            } else {
                /* Parent side. */
                /* (why is MORE executed in the parent side? Because the
                 * grandfather wants to waitpid() for the last son or
                 * grandson, and the MORE will end after the list/title
                 * command). */
                close (0);
                dup (fd[0]);
                close (fd[1]);
                if (strchr (outputname, ' ') != NULL) {
                    /* If there is a space in the command, I assume that it
                     * means that the command has arguments, so we need to
                     * call sh to split it up.
                     * We could almost equally use system() */
                    if (execl ("/bin/sh", "/bin/sh", "-c", outputname, NULL) == -1) {
                        perror ("/bin/sh");
                        exit (1);
                    }
                } else {
                    if (execlp (outputname, outputname, NULL) == -1) {
                        perror (outputname);
                        exit (1);
                    }
                }
            }
        }
        /* MUST NEVER BE REACHED !! */
        fprintf (stderr, "Murphy said: Unreachable code will be reached\n");
    } else {
        /* Parent side. */
#ifdef NO_WAITPID
        int child;
        do {
            child = wait (NULL);
            if (child == -1)
              break; /* To be honest, it's an error. */
        } while (child != pid);
#else /* have waitpid */
        waitpid (pid, NULL, 0);
#endif
    }
}
