#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include "moviedb.h"
#include "dbutils.h"            /* only for caseCompare */
#include "atree.h"
#include "readname.h"
#include "command.h"
#include "constants.h"
#include "y-or-n.h"
#include "rectify.h"
#include <readline/readline.h>

static const char *rcsid = "$Id: templedit.c,v 1.8 1994/07/20 14:04:17 queinnec Exp queinnec $";

#if 0
int fprintf (FILE *, const char *, ...);
int printf (const char *, ...);
void perror (const char *);
int fputs (const char *s, FILE *);
int fclose (FILE *);
#endif

static FILE *infile = NULL;
static FILE *outfile = NULL;
static int linenum = 0; /* input file */

typedef struct action {
    const char *keyword;        /* Producer, Genre and so on */
    void      (*read) (const struct action *); /* associated action */
    Tree       *completion_tree;        /* on which tree should we complete? */
    char     *(*complete) (char *text, int state); /* and how? */
    const char *wordbreak;      /* word separators for completion */
    int         bind_dolist;    /* M-h calls `list' on current input?  */
    int         print_info;     /* when copying to the output file, should we
                                 * also print the text?  */
} action;

static void readandprint (int print)
{
    char line[100];
    while (1) {
        if (fgets (line, 100, infile) == NULL)
          return;
        linenum++;
        if (line[0] == '\n')
          break;
        if (print)
          printf ("%s", line);
        fputs (line, outfile);
    }
    return;
}



static int readfirstline (const action *act)
{
    char line[100];
    int missing;
    if (fgets (line, 100, infile) == NULL) {
        fprintf (stderr, "Truncated file line %d\n", linenum);
        return (-1);
    }
    linenum++;
    if (line[0] == '\n') {
        fprintf (stderr, "Illegal empty line %d\n", linenum);
        return (-1);
    }
    fputs (line, outfile);
    missing = (line[0] == '#');
    if (! missing) {
        printf ("%s:\n", act->keyword);
    }
    printf ("%s", line);
    return (missing);
}

static void readnewname (const action *act, const char *actionprompt)
{
    char prompt[50];
    char *linein;
    char *rectifiedname;
    char *cleanedname;
    char *attr;
    char *charname;
    int nameisok;
    config_readline (act->completion_tree, act->complete,
                     act->wordbreak, act->bind_dolist);
    while (1) {
        sprintf (prompt, "%s %s (empty to end): ", actionprompt, act->keyword);
        linein = readline (prompt);
        if ((linein == NULL) || (*linein == '\0'))
          break;
        if (linein[strlen(linein) - 1] == ' ')
          linein[strlen(linein) - 1] = '\0';
        add_history (linein);
        rectify (linein, &cleanedname, &rectifiedname, &attr, &charname);
        free (linein);
        if (act->completion_tree) {
            /* Check that the name is in the database */
            nameisok = tree_check (act->completion_tree, cleanedname);
            if (! nameisok) {
                nameisok = y_or_n_p ("Not in the database. Confirm");
                if (nameisok)
                  tree_insert (act->completion_tree, cleanedname);
            }
        } else {
            nameisok = TRUE;
        }
        if (nameisok) {
            fprintf (outfile, "  %s", rectifiedname);
            if (attr != NULL)
              fprintf (outfile, " %s", attr);
            if (charname != NULL)
              fprintf (outfile, " ...... %s", charname);
            fprintf (outfile, "\n");
        }
        free (cleanedname);
        free (rectifiedname);
        if (attr != NULL)
          free (attr);
        if (charname != NULL)
          free (charname);
    }
    fputs ("\n", outfile);
}

static void do_name (const action *act)
{
    int missing;
    missing = readfirstline (act);
    if (missing == -1)
      return;
    readandprint (TRUE);
    if (missing)
      readnewname (act, "Enter");
    else
      readnewname (act, "Add");
    printf ("\n");
}

/* How can I ignore ..character? */
static void do_cast (const action *act)
{
    int missing;
    missing = readfirstline (act);
    if (missing == -1)
      return;
    readandprint (TRUE);
    if (missing)
      readnewname (act, "Enter new");
    else
      readnewname (act, "Add");
    printf ("\n");
}



static void do_genres (const action *act)
{
    int missing;
    missing = readfirstline (act);
    if (missing == -1)
      return;
    readandprint (TRUE);
    if (missing) {
        char *linein;
        config_readline (act->completion_tree, act->complete,
                         act->wordbreak, act->bind_dolist);
        linein = readline ("Genres: ");
        if (linein == NULL)
          return;
        if (linein[strlen(linein) - 1] == ' ')
          linein[strlen(linein) - 1] = '\0';
        if (*linein != '\0')
          fprintf (outfile, "  %s\n", linein);
        free (linein);
    }
    fputs ("\n", outfile);
    printf ("\n");
}



static void addoneline (const action *act)
{
    char prompt[70];
    char *linein;
    config_readline (NULL, NULL, NULL, FALSE);
    sprintf (prompt, "%s (empty for none): ", act->keyword);
    linein = readline (prompt);
    if (linein == NULL)
      return;
    if (*linein == '\0') {
        free (linein);
    } else {
        fprintf (outfile, "  %s\n", linein);
        add_history (linein);
    }
    fprintf (outfile, "\n");
    printf ("\n");
}

static void do_addone (const action *act)
{
    readandprint (TRUE);
    addoneline (act);
}



static void do_copy (const action *act)
{
    if (act->print_info)
      printf ("%s:\n", act->keyword);
    readandprint (act->print_info);
    fprintf (outfile, "\n");
    if (act->print_info)
      printf ("\n");
}



#define NAB FALSE /* Not applyable boolean */
#define NAP NULL  /* Not applyable pointer */

/*
 * DO NOT CHANGE THE ORDER WITHOUT CHANGING init_actions!!
 *
 * (sorry, you need a wide screen to read the following...)
 */
action the_actions[] = {
    { "Title",                  do_copy,  NAP,  NAP,          NAP,   NAB,  TRUE },
    { "Summary",                do_copy,  NAP,  NAP,          NAP,   NAB,  TRUE },
    { "Summary by",             do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Genres",                 do_genres,NULL, treecomplete, " \t,",FALSE,NAB  },
    { "Producer",               do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Producers",              do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Director",               do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Directors",              do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Writer",                 do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Writers",                do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Composer",               do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Composers",              do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Director of Photography",do_name,  NULL, treecomplete, NULL,  TRUE, NAB  }, /* V2.9 */
    { "Directors of Photography",do_name, NULL, treecomplete, NULL,  TRUE, NAB  }, /* V2.9 */
    { "Cinematographer",        do_name,  NULL, treecomplete, NULL,  TRUE, NAB  }, /* V3.0 */
    { "Cinematographers",       do_name,  NULL, treecomplete, NULL,  TRUE, NAB  }, /* V3.0 */
    { "Editor",                 do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Editors",                do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Production Designer",    do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Production Designers",   do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Costume Designer",       do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Costume Designers",      do_name,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Cast",                   do_cast,  NULL, treecomplete, NULL,  TRUE, NAB  },
    { "Cast verified by",       do_addone,NULL, NULL,         NULL,  FALSE,NAB  },
    { "Cast completed by",      do_addone,NULL, NULL,         NULL,  FALSE,NAB  },
    { "Trivia",                 do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Crazy Credits",          do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Goofs",                  do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Soundtrack",             do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Notes",                  do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Comments",               do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Corrections",            do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Quotes",                 do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Aka titles",             do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { "Running Time",           do_copy,  NAP,  NAP,          NAP,   NAB,  FALSE },
    { NULL,                     NAP,      NAP,  NAP,          NAP,   NAB,  FALSE }
};

static const char *Genres[] = {
    "Drama", "Comedy", "War", "Crime", "Musical", "Romance",
    "Biography", "Adventure", "Horror", "Thriller", "Historical",
    "Mystery", "Sci-Fi", "Fantasy", "Sports", "Spy", "Childrens",
    "Political", "Animated", "Action", "Prison", "Religious", "Opera",
    "Disaster", "Dance", "Erotic", NULL
};



static Tree *load_in_tree (const char *stem)
{
    Tree *thetree;
    thetree = tree_create (TRUE);
    if (thetree == NULL)
      exit (1);
    if (readname_dbs (thetree, stem) == ERROR) {
        fprintf (stderr, "%s not loaded\n", stem);
        exit (1);
    }
    return (thetree);
}

static void init_actions (void)
{
    Tree *atree;
    const char **w;
    
    atree = tree_create (TRUE); /* genres */
    for (w = Genres; *w != NULL; w++) {
        tree_insert (atree, *w);
    }
    the_actions[3].completion_tree = atree;

    the_actions[4].completion_tree = load_in_tree (PRDCRSTEM); /* producer */
    the_actions[5].completion_tree = the_actions[4].completion_tree; /* producers */
   
    the_actions[6].completion_tree = load_in_tree (DIRSTEM); /* director */
    the_actions[7].completion_tree = the_actions[6].completion_tree; /* directors */

    the_actions[8].completion_tree = load_in_tree (WRITESTEM); /* writer */
    the_actions[9].completion_tree = the_actions[8].completion_tree; /* writers */

    the_actions[10].completion_tree = load_in_tree (COMPSTEM); /* composer */
    the_actions[11].completion_tree = the_actions[10].completion_tree; /* composers */

    the_actions[12].completion_tree = load_in_tree (CINESTEM); /* director of photography */
    the_actions[13].completion_tree = the_actions[12].completion_tree; /* directors of photography */
    the_actions[14].completion_tree = the_actions[12].completion_tree; /* cinematographer */
    the_actions[15].completion_tree = the_actions[12].completion_tree; /* cinematographers */

    the_actions[16].completion_tree = load_in_tree (EDITSTEM); /* editor */
    the_actions[17].completion_tree = the_actions[16].completion_tree; /* editors */

    the_actions[18].completion_tree = load_in_tree (PDESSTEM); /* production designer */
    the_actions[19].completion_tree = the_actions[18].completion_tree; /* production designers */

    the_actions[20].completion_tree = load_in_tree (CDESSTEM); /* costume designer */
    the_actions[21].completion_tree = the_actions[20].completion_tree; /* costume designers */

#ifndef NOCASTLOAD
    atree = load_in_tree (ACRSTEM); /* cast */
    if (readname_dbs (atree, ACSSTEM) == ERROR) {
        fprintf (stderr, "%s not loaded\n", ACSSTEM);
        exit (1);
    }
    the_actions[22].completion_tree = atree;
#else /* NOCASTLOAD */
    the_actions[22].completion_tree = NULL;
    the_actions[22].complete = NULL;
#endif
}



static void analyse (const char *filename)
{
    char line[100];
    action *act;
    while (fgets (line, 100, infile) != NULL) {
        linenum++;
        fputs (line, outfile);
        line[strlen(line) - 1] = '\0'; /* remove trailing '\n' */
        if (line[0] == '\0')
          continue;
        if (strncmp (line, "--------", 8) == 0)
          continue;
        if (line[strlen(line) - 1] != ':') {
            fprintf (stderr, "Illegal line %d: %s\n", linenum, line);
            continue;
        }
        line[strlen(line) - 1] = '\0';
        for (act = the_actions; act->keyword != NULL; act++) {
            if (caseCompare (act->keyword, line) == 0) {
                (act->read) (act);
                break;
            }
        }
        if (act->keyword == NULL) {
            fprintf (stderr, "Illegal keyword line %d: %s\n", linenum, line);
        }
    }
}

static void launch_editor (const char *filename)
{
    const char *editorname;
    char *command;
    char *line;
    editorname = (const char *) getenv ("VISUAL");
    if (editorname == NULL)
      editorname = (const char *) getenv ("EDITOR");
    if (editorname == NULL)
      editorname = "vi";
    printf ("\n");
    printf ("Type enter to launch %s on %s,\n", editorname, filename);
    printf (" or type a full command to edit %s (including the filename),\n", filename);
    printf (" or type quit to quit.\n");
    config_readline (NULL, NULL, NULL, FALSE);
    line = readline ("-> ");
    if (line == NULL)
        return;
    if (*line == '\0') {
        execlp (editorname, editorname, filename, NULL);
        fprintf (stderr, "Can't run %s: ", editorname);
        perror ("");
        exit (1);
    } else if (strncmp (line, "quit", strlen (line)) == 0) {
        return;
    } else {
        execlp ("sh", "/bin/sh", "-c", line, NULL);
        fprintf (stderr, "Can't run sh -c %s: ", line);
        perror ("");
        exit (1);
    }
    /* NOT REACHED */
}

static void link_error (const char *f1, const char *a1, const char *f2, const char *a2)
{
    fprintf (stderr, f1, a1);
    perror ("");
    if (f2 != NULL)
      fprintf (stderr, f2, a2);
    exit (1);
}

int main (int argc, char **argv)
{
    char *infilename;
    char *outfilename;
    char *backupname;
    if (argc != 2) {
        fprintf (stderr, "templedit <filename>\n");
        exit (1);
    }
    infilename = argv[1];
    infile = fopen (infilename, "r");
    if (infile == NULL) {
        fprintf (stderr, "Can't open infile %s: ", infilename);
        perror ("");
        exit (1);
    }
    backupname = (char *) malloc (strlen (infilename) + 4 + 1);
    sprintf (backupname, "%s.bak", infilename);

    outfilename = (char *) malloc (strlen (infilename) + 4 + 1);
    sprintf (outfilename, "%s.out", infilename);
    outfile = fopen (outfilename, "w");
    if (outfile == NULL) {
        fprintf (stderr, "Can't open outfile %s: ", outfilename);
        perror ("");
        exit (1);
    }
    init_actions ();
    printf ("\nWhile entering names, you have:\n");
    printf ("  - completion, with TAB\n");
    printf ("  - full editing of current line (a la emacs)\n");
    printf ("  - history of previous entered lines (C-p, C-n and co)\n");
    printf ("  - with Meta-h (or ESC-h), you can execute `list' on the current name.\n");
    printf ("  - with Meta-H (or ESC-H), you can execute `ltrace' on the current name.\n");
    printf ("\n");
    printf ("ATTENTION: names must be entered as Hitchcock, Alfred\n");
    printf ("\n");
    analyse (infilename);
    fclose (infile);
    fclose (outfile);
    if ((unlink (backupname) == -1) && (errno != ENOENT)) {
        link_error ("Can't unlink old backup file %s: ", backupname,
                    "Edited file left in %s", outfilename);
    }
    if (link (infilename, backupname) == -1) {
        link_error ("Can't make link to backup file %s: ", backupname,
                    "Edited file left in %s", outfilename);
    }
    if (unlink (infilename) == -1) {
        link_error ("Can't cleanup temporary file %s: ",  infilename,
                    "Edited file left in %s", outfilename);
    }
    if (link (outfilename, infilename) == -1) {
        link_error ("Can't move temporary file into %s: ", infilename,
                    "Edited file left in %s", outfilename);
    }
    if (unlink (outfilename) == -1) {
        link_error ("Can't remove temporary file %s: ", outfilename,
                    NULL, NULL);
    }
    launch_editor (infilename);
    exit (0);
    return (0);
}

